{---------------------------------------------------------------------

  RZNEdit, a component with edit controls for

    real numbers      -> TRealEdit

    integers          -> TIntegerEdit

    and

    positive integers -> TCardinalEdit

  Version 1.4

  This is my version of a numeric input component. It comes with three
  types, each handling its special input range ( real, integer and
  cardinal ).
  All input is watched, a TIntegerEdit accepts only characters in
  the set [-,0,1,..,9] for instance.
  The clipboard is supported.
  An empty input is interpreted as 0.
  For TRealEdit and TIntegerEdit an input containing only a plus-sign
  or a minus-sign is also regarded as 0.
  Ancestor class is the TCustomEdit, new properties are:

    new:
    FormatString : is used in the Format-routine from Delphi,
                   eg '%12.5e'

    MaxValue \
              : determine the input boundaries, if Validate is True
    MinValue /
    Validate  : determines if input is checked against boundaries
    Value     : the result

    ColorNormal  : Color of normal text; on entering the control,
                   the color is reset to ColorNormal
    ColorWarning : Color of text, if value is not within boundaries,
                   the color is set on leaving the control's focus


  The windows setting should use a . as decimal separator


  rk 01.09.98

  Robert Kratz                             R.Kratz@fz-rossendorf.de

 ---------------------------------------------------------------------}


Unit RZNEdit;


Interface

{---------------------------------------------------------------------}

Uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, Clipbrd;

{---------------------------------------------------------------------}

Type

  TAlignment = ( AlignLeft, AlignRight );

  TRealEdit = Class( TCustomEdit )
  Private
    FFormatString, VText : String;
    FMinValue, FMaxValue : Extended;
    FValidate            : Boolean;
    FAlignment           : TAlignment;
    FColorNormal,
    FColorWarning        : TColor;

    Procedure SetValue( NewValue : Extended );
    Function  GetValue : Extended;
    Function  CheckValue( NewValue : Extended ) : Extended;

    Procedure SetValidate( B :  Boolean );
    Procedure SetFormatString( A : String );

    Function  SyntaxOk : Boolean;
    Function  EmptyOrSign : Boolean;

    Procedure CMEnter( Var Message : TCMGotFocus ); Message CM_ENTER;
    Procedure CMExit( Var Message : TCMExit );   Message CM_EXIT;

    Procedure CreateParams( Var Params : TCreateParams ); Override;

  Protected
    Function  IsValidChar( Key : Char) : Boolean; Virtual;
    Procedure KeyDown( Var Key : Word; Shift : TShiftState ); Override;
    Procedure KeyPress( Var Key : Char ); Override;
  Public
    Constructor Create( AOwner : TComponent ); Override;
  Published
    Property Alignment : TAlignment Read FAlignment Write FAlignment;
    Property AutoSelect;
    Property AutoSize;
    Property BorderStyle;
    Property Color;
    Property ColorNormal  : TColor Read FColorNormal Write FColorNormal;
    Property ColorWarning : TColor Read FColorWarning Write FColorWarning;
    Property Ctl3D;
    Property Cursor;
    Property DragCursor;
    Property DragMode;
    Property Enabled;
    Property Font;
    Property FormatString : String Read  FFormatString
                                   Write SetFormatString;
    Property Height;
    Property HelpContext;
    Property Hint;
    Property Left;
    Property MaxLength;
    Property MaxValue : Extended Read FMaxValue Write FMaxValue;
    Property MinValue : Extended Read FMinValue Write FMinValue;
    Property Name;
    Property ParentColor;
    Property ParentCtl3D;
    Property ParentFont;
    Property ParentShowHint;
    Property PopupMenu;
    Property ReadOnly;
    Property ShowHint;
    Property TabOrder;
    Property TabStop;
    Property Tag;
    Property Validate : Boolean Read FValidate Write SetValidate;
    Property Value : Extended Read GetValue Write SetValue;
    Property Visible;
    Property OnChange;
    Property OnClick;
    Property OnDblClick;
    Property OnDragDrop;
    Property OnDragOver;
    Property OnEndDrag;
    Property OnEnter;
    Property OnExit;
    Property OnKeyDown;
    Property OnKeyPress;
    Property OnKeyUp;
    Property OnMouseDown;
    Property OnMouseMove;
    Property OnMouseUp;
  End;

{---------------------------------------------------------------------}

  TIntegerEdit = Class(TCustomEdit)
  Private
    FFormatString, VText : String;
    FMinValue, FMaxValue : Longint;
    FValidate            : Boolean;
    FAlignment           : TAlignment;
    FColorNormal,
    FColorWarning        : TColor;

    Procedure SetValue( NewValue : Longint );
    Function  GetValue : Longint;
    Function  CheckValue( NewValue : Longint ): Longint;

    Procedure SetMaxValue( NewValue : Longint );   Virtual;
    Procedure SetMinValue( NewValue : Longint );   Virtual;

    Procedure SetValidate( B :  Boolean );
    Procedure SetFormatString( A : String );

    Function  SyntaxOk : Boolean;
    Function  EmptyOrSign : Boolean;

    Procedure CMEnter( Var Message : TCMGotFocus ); Message CM_ENTER;
    Procedure CMExit( Var Message : TCMExit );   Message CM_EXIT;

    Procedure CreateParams( Var Params : TCreateParams ); Override;

  Protected
    Function  IsValidChar( Key : Char) : Boolean; Virtual;
    Procedure KeyDown( Var Key : Word; Shift : TShiftState ); Override;
    Procedure KeyPress( Var Key : Char ); Override;
  Public
    Constructor Create( AOwner : TComponent ); Override;
  Published
    Property Alignment : TAlignment Read FAlignment Write FAlignment;
    Property AutoSelect;
    Property AutoSize;
    Property BorderStyle;
    Property Color;
    Property ColorNormal  : TColor Read FColorNormal Write FColorNormal;
    Property ColorWarning : TColor Read FColorWarning Write FColorWarning;
    Property Ctl3D;
    Property Cursor;
    Property DragCursor;
    Property DragMode;
    Property Enabled;
    Property Font;
    Property FormatString : String Read  FFormatString
                                   Write SetFormatString;
    Property Height;
    Property HelpContext;
    Property Hint;
    Property Left;
    Property MaxLength;
    Property MaxValue : Longint Read FMaxValue Write SetMaxValue;
    Property MinValue : Longint Read FMinValue Write SetMinValue;
    Property Name;
    Property ParentColor;
    Property ParentCtl3D;
    Property ParentFont;
    Property ParentShowHint;
    Property PopupMenu;
    Property ReadOnly;
    Property ShowHint;
    Property TabOrder;
    Property TabStop;
    Property Tag;
    Property Validate : Boolean Read FValidate Write SetValidate;
    Property Value : Longint Read GetValue Write SetValue;
    Property Visible;
    Property OnChange;
    Property OnClick;
    Property OnDblClick;
    Property OnDragDrop;
    Property OnDragOver;
    Property OnEndDrag;
    Property OnEnter;
    Property OnExit;
    Property OnKeyDown;
    Property OnKeyPress;
    Property OnKeyUp;
    Property OnMouseDown;
    Property OnMouseMove;
    Property OnMouseUp;
  End;

{---------------------------------------------------------------------}

  TCardinalEdit = Class(TIntegerEdit)
  Private
    Procedure SetMaxValue( NewValue : Longint );   Override;
    Procedure SetMinValue( NewValue : Longint );   Override;
  Protected
    Function  IsValidChar( Key : Char) : Boolean; Override;
  End;

{---------------------------------------------------------------------}
Procedure Register;
{---------------------------------------------------------------------}


{=====================================================================}


Implementation

{---------------------------------------------------------------------}

Uses WinProcs, DsgnIntf;

{---------------------------------------------------------------------}
Procedure TRealEdit.SetValue( NewValue : Extended );
{---------------------------------------------------------------------}
Begin
  Try
    Text := Format( FFormatString, [CheckValue( NewValue )] );
  Except
  End;
End;

{---------------------------------------------------------------------}
Function TRealEdit.GetValue : Extended;
{---------------------------------------------------------------------}
Begin
  Try
    If Length( Text ) <> 0 Then Result := StrToFloat( Text )
                           Else Result := 0;
  Except
    Result := FMinValue;
  End;
End;

{---------------------------------------------------------------------}
Function TRealEdit.CheckValue( NewValue : Extended ): Extended;
{---------------------------------------------------------------------}
Begin
  Result := NewValue;
  Font.Color := FColorNormal;
  If FValidate Then Begin
    If NewValue < FMinValue Then Begin
      Result:= FMinValue;
      Font.Color := FColorWarning;
    End
    Else
      If NewValue > FMaxValue Then Begin
        Result:= FMaxValue;
        Font.Color := FColorWarning;
      End;
  End;
End;

{---------------------------------------------------------------------}
Procedure TRealEdit.SetValidate( B :  Boolean );
{---------------------------------------------------------------------}
Begin
  FValidate := B;
  SetValue( Value );
End;

{---------------------------------------------------------------------}
Procedure TRealEdit.SetFormatString( A : String );
{---------------------------------------------------------------------}
Begin
  FFormatString := A;
  SetValue( Value );
End;

{---------------------------------------------------------------------}
Function TRealEdit.SyntaxOk : Boolean;
{---------------------------------------------------------------------}
Begin
  Try
    {-----------------------------------------------------------
      Syntax correct if transformation causes no exception.
      If the input is empty or a sign ( VText = '', '+', '-' ),
      it is interpreted as 0. During development within the IDE
      you may run into an extra exception.
     -----------------------------------------------------------}

    If Not EmptyOrSign Then StrToFloat( VText );
    Result := True;

  Except
    On EConvertError Do Result := False;
  End;
End;

{---------------------------------------------------------------------}
Function TRealEdit.EmptyOrSign : Boolean;
{---------------------------------------------------------------------}
Begin
  Result := (VText='') Or (VText='-') Or (VText='+');
End;

{---------------------------------------------------------------------}
Procedure TRealEdit.CMEnter( Var Message : TCMGotFocus );
{---------------------------------------------------------------------}
Begin
  If AutoSelect And Not( csLButtonDown In ControlState ) Then SelectAll;
  Font.Color := FColorNormal;
  Inherited;
End;

{---------------------------------------------------------------------}
Procedure TRealEdit.CMExit( Var Message : TCMExit );
{---------------------------------------------------------------------}
Begin
  Inherited;
  SetValue( Value );
End;

{---------------------------------------------------------------------}
Procedure TRealEdit.CreateParams( Var Params : TCreateParams );
{---------------------------------------------------------------------}
Begin
  Inherited CreateParams( Params );
  If Alignment = AlignRight Then
    Params.Style := Params.Style Or ES_MULTILINE Or ES_RIGHT;
End;

{---------------------------------------------------------------------}
Function TRealEdit.IsValidChar( Key : Char ) : Boolean;
{---------------------------------------------------------------------}
Begin
  Case Key Of
    '0'..'9', '+', '-', 'E', 'e', '.' :
       Begin
         Vtext := Copy( Text, 1, SelStart ) + Key +
                  Copy( Text, SelStart+1+SelLength, 255 );
         Result := SyntaxOk;
       End;

    #8 : Begin
           If SelLength = 0 Then
             Vtext := Copy( Text, 1, SelStart-1 ) +
                      Copy( Text, SelStart+1, 255 )
           Else
             Vtext := Copy( Text, 1, SelStart ) +
                      Copy( Text, SelStart+1+SelLength, 255 );
           Result := SyntaxOk;
         End;

  Else
    Result := False;
  End;
End;

{---------------------------------------------------------------------}
Procedure TRealEdit.KeyDown( Var Key : Word; Shift : TShiftState );
{---------------------------------------------------------------------}
Begin
  If ssCtrl In Shift Then Begin
    Case Key Of
      88 : { Ctrl X }
           If SelLength <> 0 Then Begin
             Vtext := Copy( Text, 1, SelStart ) +
                      Copy( Text, SelStart+1+SelLength, 255 );
             If SyntaxOk Then Begin
               Clipboard.AsText := SelText;
               Key := VK_DELETE;
               KeyDown( Key, [] );
             End
             Else Key := 0;
           End;

      67 : { Ctrl C }
           Clipboard.AsText := SelText;

      86 : { Ctrl V }
           If Clipboard.HasFormat(CF_TEXT) Then Begin
             Vtext := Copy( Text, 1, SelStart ) +
                      Clipboard.AsText +
                      Copy( Text, SelStart+1, 255 );
             If SyntaxOk Then Begin
               SelText := Clipboard.AsText;
               Key := 0;
             End;
           End
           Else
             MessageDlg( 'No text on the Clipboard', mtInformation,
                         [mbOK],0 );

    Else End;
  End
  Else
    If Key = VK_DELETE Then Begin
      If SelLength = 0 Then
        Vtext := Copy( Text, 1, SelStart ) +
                 Copy( Text, SelStart+2, 255 )
      Else
        Vtext := Copy( Text, 1, SelStart ) +
                 Copy( Text, SelStart+1+SelLength, 255 );
      If Not SyntaxOk Then Key := 0;
    End;

  Inherited KeyDown( Key, Shift );

End;

{---------------------------------------------------------------------}
Procedure TRealEdit.KeyPress( Var Key : Char );
{---------------------------------------------------------------------}
Begin
  If IsValidChar( Key ) Then Inherited KeyPress( Key )
                        Else Key := #0;
End;

{---------------------------------------------------------------------}
Constructor TRealEdit.Create( AOwner : TComponent );
{---------------------------------------------------------------------}
Begin
  Inherited Create( AOwner );
  CharCase := ecUpperCase;
  FMinValue := 0;
  FMaxValue := 0;
  FValidate := False;
  FAlignment := AlignLeft;
  FFormatString := '%g';
  FColorNormal  := clBlack;
  FColorWarning := clRed;
  SetValue( 0 );
End;


{=====================================================================}


{---------------------------------------------------------------------}
Procedure TIntegerEdit.SetValue( NewValue : Longint );
{---------------------------------------------------------------------}
Begin
  Try
    Text := Format( FFormatString, [CheckValue( NewValue )] );
  Except
  End;
End;

{---------------------------------------------------------------------}
Function TIntegerEdit.GetValue : Longint;
{---------------------------------------------------------------------}
Begin
  Try
    If Length( Text ) <> 0 Then Result := StrToInt( Text )
                           Else Result := 0;
  Except
    Result := FMinValue;
  End;
End;

{---------------------------------------------------------------------}
Function TIntegerEdit.CheckValue( NewValue : Longint ): Longint;
{---------------------------------------------------------------------}
Begin
  Result := NewValue;
  Font.Color := FColorNormal;
  If FValidate Then Begin
    If NewValue < FMinValue Then Begin
      Result:= FMinValue;
      Font.Color := FColorWarning;
    End
    Else
      If NewValue > FMaxValue Then Begin
        Result:= FMaxValue;
        Font.Color := FColorWarning;
      End;
  End;
End;

{---------------------------------------------------------------------}
Procedure TIntegerEdit.SetMaxValue( NewValue : Longint );
{---------------------------------------------------------------------}
Begin FMaxValue := NewValue; End;

{---------------------------------------------------------------------}
Procedure TIntegerEdit.SetMinValue( NewValue : Longint );
{---------------------------------------------------------------------}
Begin FMinValue := NewValue; End;

{---------------------------------------------------------------------}
Procedure TIntegerEdit.SetValidate( B :  Boolean );
{---------------------------------------------------------------------}
Begin
  FValidate := B;
  SetValue( Value );
End;

{------------------------------------------------------------------------------}
Procedure TIntegerEdit.SetFormatString( A : String );
{------------------------------------------------------------------------------}
Begin
  FFormatString := A;
  SetValue( Value );
End;

{---------------------------------------------------------------------}
Function TIntegerEdit.SyntaxOk : Boolean;
{---------------------------------------------------------------------}
Begin
  Try
    {-----------------------------------------------------------
      Syntax correct if transformation causes no exception.
      If the input is empty or a sign ( VText = '', '+', '-' ),
      it is interpreted as 0. During development within the IDE
      you may run into an extra exception.
     -----------------------------------------------------------}

    If Not EmptyOrSign Then StrToInt( VText );
    Result := True;
  Except
    On EConvertError Do Result := False;
  End;
End;

{---------------------------------------------------------------------}
Function TIntegerEdit.EmptyOrSign : Boolean;
{---------------------------------------------------------------------}
Begin
  Result := (VText='') Or (VText='-') Or (VText='+');
End;

{------------------------------------------------------------------------------}
Procedure TIntegerEdit.CMEnter( Var Message : TCMGotFocus );
{------------------------------------------------------------------------------}
Begin
  If AutoSelect And Not( csLButtonDown In ControlState ) Then SelectAll;
  Font.Color := FColorNormal;
  Inherited;
End;

{---------------------------------------------------------------------}
Procedure TIntegerEdit.CMExit( Var Message : TCMExit );
{---------------------------------------------------------------------}
Begin
  Inherited;
  SetValue( Value );
End;

{---------------------------------------------------------------------}
Procedure TIntegerEdit.CreateParams( Var Params : TCreateParams );
{---------------------------------------------------------------------}
Begin
  Inherited CreateParams( Params );
  If Alignment = AlignRight Then
    Params.Style := Params.Style Or ES_MULTILINE Or ES_RIGHT;
End;

{---------------------------------------------------------------------}
Function TIntegerEdit.IsValidChar( Key : Char ) : Boolean;
{---------------------------------------------------------------------}
Begin
  Case Key Of
    '0'..'9', '+', '-' :
       Begin
         Vtext := Copy( Text, 1, SelStart ) + Key +
                  Copy( Text, SelStart+1+SelLength, 255 );
         Result := SyntaxOk;
       End;

    #8 : Begin
           If SelLength = 0 Then
             Vtext := Copy( Text, 1, SelStart-1 ) +
                      Copy( Text, SelStart+1, 255 )
           Else
             Vtext := Copy( Text, 1, SelStart ) +
                      Copy( Text, SelStart+1+SelLength, 255 );
           Result := SyntaxOk;
         End;

  Else
    Result := False;
  End;
End;

{---------------------------------------------------------------------}
Procedure TIntegerEdit.KeyDown( Var Key : Word; Shift : TShiftState );
{---------------------------------------------------------------------}
Begin
  If ssCtrl In Shift Then Begin
    Case Key Of
      88 : { Ctrl X }
           If SelLength <> 0 Then Begin
             Vtext := Copy( Text, 1, SelStart ) +
                      Copy( Text, SelStart+1+SelLength, 255 );
             If SyntaxOk Then Begin
               Clipboard.AsText := SelText;
               Key := VK_DELETE;
               KeyDown( Key, [] );
             End
             Else Key := 0;
           End;

      67 : { Ctrl C }
           Clipboard.AsText := SelText;

      86 : { Ctrl V }
           If Clipboard.HasFormat(CF_TEXT) Then Begin
             Vtext := Copy( Text, 1, SelStart ) +
                      Clipboard.AsText +
                      Copy( Text, SelStart+1, 255 );
             If SyntaxOk Then Begin
               SelText := Clipboard.AsText;
               Key := 0;
             End;
           End
           Else
             MessageDlg( 'No text on the Clipboard', mtInformation,
                         [mbOK],0 );

    Else End;
  End
  Else
    If Key = VK_DELETE Then Begin
      If SelLength = 0 Then
        Vtext := Copy( Text, 1, SelStart ) +
                 Copy( Text, SelStart+2, 255 )
      Else
        Vtext := Copy( Text, 1, SelStart ) +
                 Copy( Text, SelStart+1+SelLength, 255 );
      If Not SyntaxOk Then Key := 0;
    End;

  Inherited KeyDown( Key, Shift );

End;

{---------------------------------------------------------------------}
Procedure TIntegerEdit.KeyPress( Var Key : Char );
{---------------------------------------------------------------------}
Begin
  If IsValidChar( Key ) Then Inherited KeyPress( Key )
                        Else Key := #0;
End;

{---------------------------------------------------------------------}
Constructor TIntegerEdit.Create( AOwner : TComponent );
{---------------------------------------------------------------------}
Begin
  Inherited Create( AOwner );
  FMinValue := 0;
  FMaxValue := 0;
  FValidate := False;
  FAlignment := AlignLeft;
  FFormatString := '%d';
  FColorNormal  := clBlack;
  FColorWarning := clRed;
  SetValue( 0 );
End;


{=====================================================================}

{---------------------------------------------------------------------}
Procedure TCardinalEdit.SetMaxValue( NewValue : Longint );
{---------------------------------------------------------------------}
Begin
  If NewValue<0 Then NewValue:=0;
  FMaxValue := NewValue;
End;

{---------------------------------------------------------------------}
Procedure TCardinalEdit.SetMinValue( NewValue : Longint );
{---------------------------------------------------------------------}
Begin
  If NewValue<0 Then NewValue:=0;
  FMinValue := NewValue;
End;

{---------------------------------------------------------------------}
Function TCardinalEdit.IsValidChar( Key : Char ) : Boolean;
{---------------------------------------------------------------------}
Begin
  Case Key Of
    '0'..'9' :
       Begin
         Vtext := Copy( Text, 1, SelStart ) + Key +
                  Copy( Text, SelStart+1+SelLength, 255 );
         Result := SyntaxOk;
       End;

    #8 : Begin
           If SelLength = 0 Then
             Vtext := Copy( Text, 1, SelStart-1 ) +
                      Copy( Text, SelStart+1, 255 )
           Else
             Vtext := Copy( Text, 1, SelStart ) +
                      Copy( Text, SelStart+1+SelLength, 255 );
           Result := SyntaxOk;
         End;

  Else
    Result := False;
  End;
End;

{---------------------------------------------------------------------}
Procedure Register;
{---------------------------------------------------------------------}
Begin
  RegisterComponents('MyComponents', [TRealEdit]);
  RegisterComponents('MyComponents', [TIntegerEdit]);
  RegisterComponents('MyComponents', [TCardinalEdit]);
End;

{---------------------------------------------------------------------}
End.

